#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "gluw.h"
#include "get_token.h"
#include "vector.h"
#include "obj.h"
#include "hair.h"

#define STR_COLUM_MAX    65536

int add_HG(struct HairGuide *srcHG, struct HG_List *HGL ) ;

struct HairObjData *read_hair( char *fname )
{
  FILE   *fp ;
  char strline[STR_COLUM_MAX] ;
  struct TOKEN_STRUCT tk ;
  int    i ;

  struct HairObjData *hobj ;
  struct PolyObjData *obj  ;
  struct HairGuide    HG   ;
  struct HG_List     *HGL  ;
  int fnum=-1 ;
  int pnum=-1 ;
  int pcount=0 ;
  int fcount=0 ;
  int gcount=0 ;

  hobj=init_hairobj() ;
  if( hobj==NULL ) return(NULL) ;
  
  obj=(struct PolyObjData *)malloc(sizeof(struct PolyObjData)) ;
  if( obj==NULL ){
    free_hairobj(hobj) ;
    return(NULL) ;
  }
  obj->px=0 ; obj->py=0 ; obj->pz=0 ;
  obj->ox=0 ; obj->oy=0 ; obj->oz=0 ;
  obj->drawx=0 ;
  obj->drawy=0 ;
  obj->drawz=0 ;
  obj->plist=NULL ;
  obj->pnum=0 ;
  obj->flist=NULL ;
  obj->fnum=0 ;
  
  fp=fopen(fname,"rb") ;
  if( fp==NULL ){
    free(obj) ;
    free_hairobj( hobj ) ;
    return(NULL) ;
  }

  HGL=hobj->hgl ;
  
  for( i=0 ; i<HGL->n ; i++ ){
    HGL->HG[i].pnum =0 ;
  }
  
  while( -1 ){
    if( fgets( strline , (STR_COLUM_MAX-1) , fp )==NULL ) break ;
    get_token( strline , &tk ) ;

    /** read face num **/
    if( fnum==-1 && strcmp(tk.pt[0],"FN")==0 ){
      fnum=atoi(tk.pt[1]) ;
      obj->fnum=fnum ;
      obj->flist=(struct Face3D *)malloc( sizeof(struct Face3D) * obj->fnum ) ;
      if( obj->flist==NULL ){
	if( obj->plist!=NULL ) free(obj->plist) ;
	if( obj->nlist!=NULL ) free(obj->nlist) ;
	if( obj->sel  !=NULL ) free(obj->sel  ) ;
	if( obj->draw !=NULL ) free(obj->draw ) ;
	free( obj ) ;
	free_hairobj( hobj ) ;
	return(NULL) ;
      }
      continue ;
    }
    /** read point num **/
    if( pnum==-1 && strcmp(tk.pt[0],"PN")==0 ){
    
      pnum=atoi(tk.pt[1]) ;
      obj->pnum=pnum ;
      obj->plist=(struct Vec3D *)malloc( sizeof(struct Vec3D) * obj->pnum ) ;
      obj->nlist=(struct Vec3D *)malloc( sizeof(struct Vec3D) * obj->pnum ) ;
      obj->HGid =(int*)malloc( sizeof(int) * obj->pnum ) ;
      obj->sel  =(int*)malloc( sizeof(int) * obj->pnum ) ;
      obj->draw =(int*)malloc( sizeof(int) * obj->pnum ) ;
      if( obj->plist==NULL ||
	  obj->nlist==NULL ||
	  obj->HGid ==NULL ||
	  obj->sel  ==NULL ||
      	  obj->draw ==NULL ){
	if( obj->plist==NULL ) free(obj->plist) ;
	if( obj->nlist==NULL ) free(obj->nlist) ;
	if( obj->HGid ==NULL ) free(obj->HGid)  ;
	if( obj->sel  ==NULL ) free(obj->sel)   ;
	if( obj->draw ==NULL ) free(obj->draw)  ;
	free( obj ) ;
	free_hairobj( hobj ) ;
	return(NULL) ;
      }
      for( i=0 ; i<obj->pnum ; i++ ){
	obj->nlist[i].x=0.0 ;
	obj->nlist[i].y=0.0 ;
	obj->nlist[i].z=0.0 ;
	obj->HGid[i]   =-1 ;
	obj->draw[i]   = 0 ;
	obj->sel[i]    = 0 ;
      }
      continue ;
    }
    /** read point **/
    if( pnum!=-1 && strcmp(tk.pt[0],"P")==0 ){
      int pid ;
      pid=atoi(tk.pt[1]) ;
      obj->plist[pid].x=atof(tk.pt[2]) ;
      obj->plist[pid].y=atof(tk.pt[3]) ;
      obj->plist[pid].z=atof(tk.pt[4]) ;
      pcount++ ;
      continue ;
    }
    /** read face **/
    if( fnum!=-1 && strcmp(tk.pt[0],"F")==0 ){
      int fid ;
      
      fid=atoi(tk.pt[1]) ;
      obj->flist[fid].p0=atoi(tk.pt[2]) ;
      obj->flist[fid].p1=atoi(tk.pt[3]) ;
      obj->flist[fid].p2=atoi(tk.pt[4]) ;
      fcount++ ;
      continue ;
    }
    /** read hair guide **/
    if( strcmp(tk.pt[0],"HG")==0 ){
      HG.type = atoi(tk.pt[1]) ;
      HG.gid  = atoi(tk.pt[2]) ;
      HG.pid  = atoi(tk.pt[3]) ;
      HG.pnum = atoi(tk.pt[4])*3 ;
      HG.nknot= 8 ;
      if( HG.gid<-1 || HG.gid>=GPNUM ) HG.gid=0 ;
      for( i=0 ; i<HG.nknot ; i++ ){
	if( i<HG.nknot/2 ){
	  HG.knots[i]=0.0 ;
	}else{
	  HG.knots[i]=1.0 ;
	}
      }
      for( i=0 ; i<HG.pnum ; i++ ){
	HG.plist[i]=atof(tk.pt[i+5]) ;
      }
      if( add_HG(&HG,HGL)<0 ){
	fclose(fp) ;
	return(NULL) ;
      }
      gcount++ ;
    }
  }

  fclose(fp) ;

  for( i=0 ; i<obj->fnum ; i++ ){
    struct Vec3D vec0 ;
    struct Vec3D vec1 ;

    /** calc face normals **/
    SubVec( &obj->plist[obj->flist[i].p0],
	    &obj->plist[obj->flist[i].p1],
	    &vec0 ) ;
    SubVec( &obj->plist[obj->flist[i].p0],
	    &obj->plist[obj->flist[i].p2],
	    &vec1 ) ;
    CrossProduct( &vec1, &vec0, &obj->flist[i].normal ) ;
    NormalizeVec(&obj->flist[i].normal, &obj->flist[i].normal) ;
    
    /** calc point normals **/
    AddVec( &obj->nlist[obj->flist[i].p0],
	    &obj->flist[i].normal,
	    &obj->nlist[obj->flist[i].p0] ) ;
    AddVec( &obj->nlist[obj->flist[i].p1],
	    &obj->flist[i].normal,
	    &obj->nlist[obj->flist[i].p1] ) ;
    AddVec( &obj->nlist[obj->flist[i].p2],
	    &obj->flist[i].normal,
	    &obj->nlist[obj->flist[i].p2] ) ;
  }
  for( i=0 ; i<obj->pnum ; i++ ){
    NormalizeVec( &obj->nlist[i], &obj->nlist[i] ) ;
  }

  for( i=0 ; i<HGL->n ; i++ ){ /*** $B%X%"%,%$%I9=B$BNFb$KD:E@>pJs$r%3%T!<$9$k(B ***/
    HGL->HG[i].rp= obj->plist[HGL->HG[i].pid] ;
    HGL->HG[i].rn= obj->nlist[HGL->HG[i].pid] ;
  }

  hobj->head=obj ;

  return( hobj ) ;
}
